import graph
from IPython import display

class PositionController(object):
    def __init__(self, left=0, top=0):
        self.left = left
        self.top = top
    def copy(self):
        return PositionController(self.left, self.top)

def make_arrow(x1,y1,x2,y2):
    h = '<svg height="{}" width="{}">'.format(y2 + 20,x2 + 20)
    a = '''
    <defs>
        <marker id="markerArrow" markerWidth="12" markerHeight="12" refX="2" refY="6" orient="auto">
        <path d="M2,2 L2,11 L10,6 L2,2" style="fill: #ffffff;" />
        </marker>
    </defs>
    '''
    b = '<line x1="{}" y1="{}" x2="{}" y2="{}" class="arrow" /></svg>'.format(x1,y1,x2,y2)
    return h + a + b


class VisualizationProbe(object):
    def __init__(self):
        self.document = "<html>"
        self.style = '''
        <style>
            body {
                background-color: black;
                font-family: Arial, Helvetica, sans-serif;
            }
            div {
                position: absolute;
                top: 100px;
                left: 10;
                width: fit-content;
                height: fit-content;
                border: 3px solid #73AD21;
                color: orange;
            }
            .arrow {
                stroke: rgb(255, 255, 255);
                stroke-width: 2;
                marker-end: url(#markerArrow)
            }
            svg {
                position: absolute
            }
        </style>
        '''
        self.document += self.style
    def trace(self, node, position=None):

        if position is None:
            position = PositionController()

        
        
        tag_style = "top: {}px; left: {};".format(position.top, position.left)

        div = '<div style="{}">'.format(tag_style)
        node_tag = "<h3>" + str(node).replace("<","&lt;").replace(">","&gt;") + "</h3>"
        node_tag = div + node_tag + "</div>"
        self.document += node_tag

        position_copy = position.copy()
        position_copy.top += 100
        for i, parent in enumerate(node.parents):
            arrow = make_arrow(
                position.left,
                position.top,
                position_copy.left,
                position_copy.top
            )
            self.document += arrow
            self.trace(parent, position_copy)
            position_copy.left += 400

        
        
    def display(self):
        self.document += "</html>"
        #display.HTML(self.document)
        return self.document

def summary(node):
    visualization_probe = VisualizationProbe()
    visualization_probe.trace(node)
    
    with open("graph_visualizations/out.html","w") as f:
        f.write(visualization_probe.display())



import jax.numpy as jnp
from jax import grad, jacfwd
import numpy as np

class GraphNode(object):
    def __init__(self):
        self.cache = None
        self.local_grad_cache = None
        self.global_grad_cache = None

        # There are certain types of special nodes that are very
        # important to the graph algorithm's functioning. Probes
        # will nead to know if a node is one of the special types 
        # of nodes

        # if a node is a tensor, then it does not have any parents
        # and it has a gradient variable 
        self.isTensor = False
        # if a node is a ForkNode, then that means the graph splits 
        # at it, and it needs to be treated specially
        self.isFork = False

        self.parents = list()
        self.grad_fns = list()

        self.tip = self
        self.link = self
   
    
class GraphModule(object):
    def __init__(self):
        pass

class Tensor(GraphNode):
    def __init__(self, param, frozen=False, des=""):
        super().__init__()
        self.isTensor = True
        self.param = param
        self.grad = None
        
        self.des = des
        self.frozen = frozen
    def shape(self):
        return self.param.shape
    def __str__(self):
        return "Tensor {}, shape:{}".format(self.des, self.param.shape)
    def __len__(self):
        return self.param.shape[0]
    def __getitem__(self, idx):
        return Tensor(self.param[idx], frozen=self.frozen, des=self.des)

  
        

class ForkNode(GraphNode):
    def __init__(self, parent, fork_count):
        super.__init__(self)
        self.isFork = True
        self.parents = [parent]
        self.fork_count = fork_count
        self.fork_counter_cache = 0
        self.fork_cache = None
    @staticmethod
    def fn(x):
        return x
    


def forward_trace( node ):
    if node.isTensor:
        return node.param
    else:
        node.cache = list()
        for parent in node.parents:
            node.cache.append(forward_trace(parent))
        out = node.fn(*node.cache)
        return out


def gradient_trace(node, phi=None):
    if phi is None:
        phi = lambda x : x
    for i, parent in enumerate(node.parents):
        psi = lambda y: phi(node.fn( *node.cache[:i], y, *node.cache[i+1:] ))
        if parent.isTensor:
            if not parent.frozen:
                parent.grad = grad(psi)(node.cache[i])
        gradient_trace(parent, psi)    
        

class OptimizationProbe(object):
    def __init__(self, optimizer):
        self.optimizer = optimizer
    def trace(self, node):
        # TODO
        # this algorithm needs to account for fork nodes
        if node.isTensor:
            #print(node)
            if not node.frozen:
                # the only thing a tensor does is store a param value, 
                # so its global gradient cache will only have one element
                #print(node)
                node.param = self.optimizer.optimize(node.param, node.grad)
        else:
            if node.isFork:
                # this will manage backwards differentiation for fork nodes

                # count how many times the probe has reached the fork. 
                node.fork_counter_cache += 1
                
                # if the probe has reached the fork as many times as it splits,
                # then it the probe will continue tracing
                if node.fork_counter_cache == node.fork_count:
                    self.trace(node.parents[0])
            else:
                for parent in node.parents:
                    self.trace(parent)

def clear_cache(node):
        if node.isFork:
            node.fork_cache = None
            node.fork_counter_cache = None
        if node.local_grad_cache is not None:
            node.local_grad_cache = None
        if node.global_grad_cache is not None:
            node.global_grad_cache = None
        if node.isTensor:
            if node.grad is not None:
                node.grad = None
        if node.cache is not None:
            node.cache = None
            for parent in node.parents:
                clear_cache(parent)




from PIL import Image
from numpy import asarray, frombuffer
import graph
import numpy as np
import jax.numpy as jnp

class MNIST(object):
    def __init__(self):
        image = Image.open('utils/mnist_sprite.png')
        self.data = asarray(image) / 255.0

        with open ('utils/mnist_labels_uint8','rb') as fp:
            integers = frombuffer(fp.read(65000 * 10), dtype='uint8')
        self.labels = integers.reshape(65000, 10)
    def __len__(self):
        return self.data.shape[0]
    def __getitem__(self, i):
        return graph.Tensor(jnp.array(self.data[i]), frozen=True), graph.Tensor(self.labels[i].astype(np.float64), frozen=True, des="MNIST label")



import graph
import jax.numpy as jnp
from jax import grad

class LossNode(graph.GraphNode):
    def __init__(self, parent):
        super().__init__()
        self.parents = [parent, None]

    def attach(self, label):
        self.parents[1] = label



class MeanSquaredError(LossNode):
    def __init__(self, parent):
        super().__init__(parent)
    @staticmethod
    def fn(x,y):
        return jnp.mean((x - y) ** 2)
    def __str__(self):
        return "Mean Squared Error Loss"



class CategoricalCrossEntropy(LossNode):
    def __init__(self, parent):
        super().__init__(parent)
    @staticmethod
    def fn(x,y):
        #print("input",x)
        smx = jnp.exp(x)
        #print("exp",smx)
        smx = smx / jnp.sum(smx)
        #print("smx",smx)
        return -jnp.sum(y * jnp.log(smx))
    


import numpy as np

import visual
import loss
import graph
import optim
import nn



def inference(link):
    forward_probe = graph.ForwardProbe()
    out = forward_probe.trace(link)
    forward_probe.clear_cache(link)
    return out

def optimize(link):
    forward_probe = graph.ForwardProbe()
    forward_probe.trace(link)
    grad_probe = graph.GradientProbe()
    grad_probe.trace(link, None)
    optimizer = optim.StochasticGradientDescent(lr=0.01)
    optim_probe = graph.OptimizationProbe(optimizer)
    optim_probe.trace(link)

    forward_probe.clear_cache(mse)




inp = graph.Tensor(np.array([5.0,5.0,5.0]), frozen=True, des="input")
lnr = nn.Linear(inp,[2,3])
lbl = graph.Tensor(np.array([6.0,6.0]), frozen=True, des="label")
mse = loss.MeanSquaredError(lnr.link)
mse.attach(lbl)

#visual.summary(mse)

for i in range(10):
    optimize(mse)
    print(inference(mse))



'''

forward_probe = graph.ForwardProbe()
print(forward_probe.trace(mse))

grad_probe = graph.GradientProbe()
grad_probe.trace(mse, None)

optimizer = optim.StochasticGradientDescent(0.001)
optim_probe = graph.OptimizationProbe(optimizer)
optim_probe.trace(mse)

visual.summary(mse)
'''


import graph
from jax import grad
import jax.numpy as jnp
import numpy as np

class MatMul(graph.GraphNode):
    def __init__(self, parents):
        super().__init__()
        self.parents = parents
    @staticmethod
    def fn(x,y):
        return x @ y

class Add(graph.GraphNode):
    def __init__(self, parents):
        super().__init__()
        self.parents = parents
    @staticmethod
    def fn(x,y):
        return x + y

class Weight(graph.GraphModule):
    def __init__(self, parent, shape):
        self.weight = graph.Tensor(jnp.array(np.random.rand(*shape) * 0.01), des="weight")
        #self.weight = graph.Tensor(jnp.array([[1.,2.],[1.,2.],[1.,2.]]), "weight")

        self.matmul = MatMul([self.weight, parent])

        self.link = self.matmul
    def __str__(self):
        return "Weight, dim: {}".format(self.weight.param.shape)

class Bias(graph.GraphModule):
    def __init__(self, parent, dim):
        self.bias = graph.Tensor(jnp.array(np.random.rand(dim) * 0.01), des="bias")
        #self.bias = graph.Tensor(jnp.array([1.,2.]), "bias")

        self.add = Add([parent, self.bias])

        self.link = self.add
    def __str__(self):
        return "Bias, dim: {}".format(self.bias.param.shape)

class Linear(graph.GraphModule):
    def __init__(self, parent, shape):

        self.weight = Weight(parent, shape)
        self.bias = Bias(self.weight.link, shape[0])

        self.link = self.bias.link

class Input(graph.GraphNode):
    def __init__(self):
        super().__init__()
        self.isAttached = False
    def attach(self, parent):
        self.parents = [parent]
        self.isAttached = True
    def __str__(self):
        return "Graph Input. Attached: {}".format(self.isAttached)
    @staticmethod
    def fn(x):
        return x


class ReLU(graph.GraphNode):
    def __init__(self, parent):
        super().__init__()
        self.parents = [parent]
    @staticmethod
    def fn(x):
        return jnp.maximum(x,0)

class Softmax(graph.GraphNode):
    def __init__(self, parent):
        super().__init__()
        self.parents = [parent]
    @staticmethod
    def fn(x):
        exp = jnp.exp(x)
        return exp / jnp.sum(exp)

